/*
    This file is part of Leela Zero.
    Copyright (C) 2017 Gian-Carlo Pascutto

    Leela Zero is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Leela Zero is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Leela Zero.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef RANDOM_H_INCLUDED
#define RANDOM_H_INCLUDED

#include <cstdint>
#include <limits>

/**
* @file Random.h
* @brief Header of Random.cpp
*
*/

/**
* @namespace Utils
* @brief A set of useful method and class
*
*/
namespace Utils {

    /**
    * @class Random
    * @brief RNG based on Xoroshiro128+
    *
    * @copyright GNU General Public License
    * @copyright Part of Leela Zero software
    * @author Gian-Carlo Pascutto
    */
    class Random {
    public:
        /**
        * @fn Random
        * @brief Free the pointed object
        *
        */
        Random() = delete;

        /**
        * @fn Random(std::uint64_t seed = 0)
        * @brief Initialize the random number generator with a seed
        *
        * @param seed The value of the seed
        */
        Random(std::uint64_t seed = 0);

        /**
        * @fn seedrandom(std::uint64_t s)
        * @brief Initialize state of xoroshiro128+ by transforming the seed with the splitmix64 algorithm.
        *
        * @param s The value of the seed
        */
        void seedrandom(std::uint64_t s);

        // Random numbers from [0, max - 1]
        template<int MAX>
        /**
        * @fn randfix
        * @brief Check that the random number will be generated in a correct range, then generate it
        *
        * @return Random generated number between [0, max-1]
        */
        std::uint32_t randfix() {
            static_assert(0 < MAX && MAX < std::numeric_limits<std::uint32_t>::max(),"randfix out of range");
            // Last bit isn't random, so don't use it in isolation. We specialize
            // this case.
            static_assert(MAX != 2, "don't isolate the LSB with xoroshiro128+");
            return gen() % MAX;
        }

        /**
        * @fn randuint64
        * @brief Generate raw random number
        *
        * @result The random generated number
        */
        std::uint64_t randuint64();

        // Random number from [0, max - 1]
        /**
        * @fn randuint64(const std::uint64_t max)
        * @brief Generate random number between [0, max-1] using uniform_int_distribution
        *
        * @result The random generated number
        */
        std::uint64_t randuint64(const std::uint64_t max);

        // return the thread local RNG
        /**
        * @fn get_Rng(void)
        * @brief Get a copy of the current thread
        *
        * @return A thread local RNG
        */
        static Random& get_Rng(void);

        // UniformRandomBitGenerator interface
        using result_type = std::uint64_t;

        /**
        * @fn min
        * @brief Min integer value, part of UniformRandomBitGenerator interface
        * 
        * @return Lowest uint64_t value
        */
        constexpr static result_type min() {
            return std::numeric_limits<result_type>::min();
        }

        /**
        * @fn max
        * @brief Max integer value, part of UniformRandomBitGenerator interface
        * 
        * @return Highest uint64_t value
        */
        constexpr static result_type max() {
            return std::numeric_limits<result_type>::max();
        }

        /**
        * @fn operator()()
        * @brief Generate a raw random number, part of UniformRandomBitGenerator interface
        *
        * @return The random generated number
        */
        result_type operator()() {
            return gen();
        }

    private:
        /**
        * @fn gen(void)
        * @brief Generate random generated number with Xoroshiro128+
        *
        * @return Random generated number 
        */
        std::uint64_t gen(void);

        std::uint64_t m_s[2]; /**< Array used to compute random numbers*/
    };

    // Specialization for last bit: use sign test
    template<>
    inline std::uint32_t Random::randfix<2>() {
        return (gen() > (std::numeric_limits<std::uint64_t>::max() / 2));
    }
}

#endif