/*
    This file is part of Leela Zero.
    Copyright (C) 2017 Gian-Carlo Pascutto

    Leela Zero is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Leela Zero is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Leela Zero.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SMP_H_INCLUDED
#define SMP_H_INCLUDED

#include <atomic>

/**
* @namespace SMP
* @brief A Symmetric multiprocessing implementation
*
*/
namespace SMP {
	/**
	* @fn get_num_cpus
	* @brief Returns the number of hardware thread contexts. This value is supposed to be equals to the number of cores of the machine.
	*
	* @return The number of hardware thread contexts 
	*/
    int get_num_cpus();

	/**
	* @class Mutex
	* @brief Class used to protect data shared with different threads.
	*
	* @copyright GNU General Public License
    * @copyright Part of Leela Zero software
    * @author Gian-Carlo Pascutto
	*/
    class Mutex {
    public:
		/**
		* @fn Mutex
		* @brief Standard constructor (Set the Mutex as unlocked)
		*
		*/
        Mutex();
		
		/**
		* @fn ~Mutex
		* @brief Standard destructor
		*
		*/
        ~Mutex() = default;
        friend class Lock;
    private:
        std::atomic<bool> m_lock; /**< Boolean representing the locking state of the Mutex */
    };

	/**
	* @class Lock
	* @brief Class used to handle Mutex
	*
	* @copyright GNU General Public License
    * @copyright Part of Leela Zero software
    * @author Gian-Carlo Pascutto
	*/
    class Lock {
    public:
		/**
		* @fn Lock
		* @brief Construct a Lock with a Mutex
		*
		* @param[in] m
		*/
        explicit Lock(Mutex & m);
		/**
		* @fn ~Lock
		* @brief Standard destructor
		*
		*/
        ~Lock();
		/**
		* @fn lock
		* @brief Try to lock the Mutex
		*
		*/
        void lock();
		/**
		* @fn unlock
		* @brief Unlock the Mutex
		*
		*/
        void unlock();
    private:
        Mutex * m_mutex; /**< The associated mutex*/
    };
}

// Avoids accidentally creating a temporary
#define LOCK(mutex, lock) SMP::Lock lock((mutex))

#endif
