/*
    This file is part of Leela Zero.
    Copyright (C) 2017 Gian-Carlo Pascutto

    Leela Zero is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Leela Zero is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Leela Zero.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef UTILS_H_DEFINED
#define UTILS_H_DEFINED

#include <atomic>
#include <limits>
#include <string>
#include <fstream>

#include "ThreadPool.h"
#include <vector>
#include <ale_interface.hpp>

/**
* @file Utils.hpp
* @brief Header of Utils.hpp
*
*/

using namespace std;

extern Utils::ThreadPool thread_pool;
extern ALEInterface ALE; /**< ALE interface used to run the agent */ 

/**
* @namespace Utils
* @brief A set of useful method and class
*
* @copyright GNU General Public License
* @copyright Part of Leela Zero software
* @author Gian-Carlo Pascutto
*/
namespace Utils {
    template<class T>
    /**
    * @fn atomic_add(std::atomic<T> &f, T d)
    * @brief Perform an atomic addition
    *
    * @param[in,out] f Original atomic value
    * @param d Value added to the atomic value
    */
    void atomic_add(std::atomic<T> &f, T d) {
        T old = f.load();
        while (!f.compare_exchange_weak(old, old + d));
    }

    template<typename T>
    
    /**
    * @fn rotl(const T x, const int k)
    * @brief Rotate bytes to the left
    *
    * @param x Value to be rotated
    * @param k Number of bits to shift
    * @return The rotated value
    */
    T rotl(const T x, const int k) {
        return (x << k) | (x >> (std::numeric_limits<T>::digits - k));
    }

	/**
	* @fn bool is7bit(int c)
	* @brief Check if a value is written on 7 bits
	*
	* @param c The checked value
	* @return True if the value is written on 7 bits, false otherwise 
	*/
    inline bool is7bit(int c) {
        return c >= 0 && c <= 127;
    }

	/**
	* @fn size_t Utils::ceilMultiple(size_t a, size_t b)
	* @brief Calculate the closest value from a such that this new value is a multiple of b
	*
	* @param a The divided
	* @param b The divisor
	* @return The closest value from a such that b divide this value
	*/
    size_t ceilMultiple(size_t a, size_t b);

	/**
	* @fn int Utils::secam_to_range_index(int i)
	* @brief Transform a secam color value to a power of two
	*
	* @param i The color value we want to transform
	* @return The transformed color value
	*/
    inline int secam_to_range_index(int i);

	/**
	* @fn Utils::secam210x160_to_42x32(const std::vector<unsigned char>& screen,std::vector<unsigned char>& screen42x32)
	* @brief Reduce the screen from 210x160 to 42x32
	*
	* @param[in] screen The input screen (pointer vector)
	* @param[out] screen42x32 The output screen (pointer vector)
	* @see secam_to_range_index()
	*/
    void secam210x160_to_42x32(const std::vector<unsigned char>& screen,
                               std::vector<unsigned char>& screen42x32);

	/**
	* @fn init_ALE()
	* @brief Initialize the emulator from the config file
	* @details The properties set are repeat_action_probability, display_screen, sound, ROM and colourPalette
	*
    */
    void init_ALE();

    /**
	* @fn read_Config()
	* @brief Store values extracted from the config file to a vector
	* @details The config file is config.config. This file have to be flawlessly written or the algorithm will crash.
    *
    * @return A vector of vector of string. All the lines of the file are stored in a vector. Lines are represented with a vector containing a parameter and his value
    */
    vector<vector<string>> read_Config();
}

#endif

