#include "agent.hpp"
#include "Random.h"
#include <random>
#include <algorithm>

/**
* @file agent.cpp
* @brief Handle the true representation of an agent, i.e. an AI.
*
*/

agent::agent(const parameters& params)
  : params(params),
    program(params.program_size * params.nb_parts),
    registers(params.nb_registers + params.nb_long_term_registers + params.output_size)
{
}

void agent::reset()
{
  std::fill(begin(registers), end(registers), 0);
  nb_steps = 0;
  current_part = 0;
}

void agent::reset_registers()
{
  std::fill(begin(registers), begin(registers) + params.nb_registers + params.output_size, 0);
}

void agent::execute_instruction(const std::vector<unsigned char>& screen, int& pc)
{
  instruction ins = program[pc];
   //print_instruction(std::cout, ins);
  // std::cout << std::endl;
  int opcode = ins & 0xf;
  if (opcode == OP_IF)
    {
      int reg3 = (ins >> 41) & 0x7ff;
      double op3  = reg3 < params.input_size ? screen[reg3] : registers[reg3 - params.input_size];
      int reg4 = (ins >> 52) & 0x7ff;
      double op4  = reg4 < params.input_size ? screen[reg4] : registers[reg4 - params.input_size];
      if (op3 >= op4) return;
      opcode = (ins >> 37) & 0xf;
    }
  int dest = ((ins >> 4) & 0x7ff) - params.input_size;
  //  std::cout << dest << std::endl;
  int reg1 = (ins >> 15) & 0x7ff;
  double op1  = reg1 < params.input_size ? screen[reg1] : registers[reg1 - params.input_size];
  //std::cout << reg1 << " " << op1 << std::endl;
  int reg2 = (ins >> 26) & 0x7ff;
  double op2  = reg2 < params.input_size ? screen[reg2] : registers[reg2 - params.input_size];
  //std::cout << reg2 << " " << op2 << std::endl;
  switch (opcode)
    {
    case OP_ADD:
      registers[dest] = op1 + op2;
      break;

    case OP_MINUS:
      registers[dest] = op1 - op2;
      break;

    case OP_MAX:
      registers[dest] = std::max(op1, op2);
      break;

    case OP_MIN:
      registers[dest] = std::min(op1, op2);
      break;

    case OP_UMINUS:
      registers[dest] = - op1;
      break;

    case OP_MUL:
      registers[dest] = op1 * op2;
      break;

    case OP_DIV:
      registers[dest] = op2 == 0 ? 0 : op1 / op2;
      break;

    case OP_LOG:
      registers[dest] = op1 <= 0 ? 0 : std::log(op1);
      break;

    case OP_EXP:
      registers[dest] = std::exp(op1);
      break;

    case OP_COS:
      registers[dest] = std::cos(op1);
      break;

    case OP_SIN:
      registers[dest] = std::sin(op1);
      break;

    case OP_POPCOUNT:
      registers[dest] = __builtin_popcount(op1);
      break;

    case OP_IF_GOTO:
      {
      int reg3 = (ins >> 41) & 0x7ff;
      double op3  = reg3 < params.input_size ? screen[reg3] : registers[reg3 - params.input_size];
      if (op1 < op2)
        {
          int old_pc = pc;
          pc = pc + static_cast<int>(op3);
          if (pc < old_pc) pc = old_pc;
        }
      break;
      }
    default:
      break;
    }
}

auto agent::begin_outputs()
{
  return begin(registers) + params.nb_registers;
}

auto agent::end_outputs()
{
  return begin(registers) + params.nb_registers + params.output_size;
}

int agent::get_action(const std::vector<unsigned char>& screen)
{
  reset_registers();
  int start = current_part * params.program_size;
  if (current_part < params.nb_parts - 1)
    {
      if (nb_steps >= params.time_by_part)
        {
          nb_steps = 0;
          current_part++;
        }
    }
  for (int pc = start; pc < start + params.program_size; ++pc)
    {
      execute_instruction(screen, pc);
    }
  ++nb_steps;
  // std::uniform_real_distribution<double> r(0, 1e-9);
  // auto it = begin_outputs();
  // for (int i = 0; i < params.output_size; ++i)
  //   {
  //     *it++ += r(Utils::Random::get_Rng());
  //   }
  // for (auto it = begin_outputs(); it != end_outputs(); ++it)
  //   {
  //     std::cout << *it << " ";
  //   }
  // for (int i = 0; i < params.output_size + params.nb_registers + params.nb_blocks; ++i)
  //   {
  //     std::cout << registers[i] << " ";
  //   }
  // std::cout << std::endl;
  return std::max_element(begin_outputs(), end_outputs()) - begin_outputs();
}

std::string agent::reg_to_string(int reg) const
{
  if (reg < params.input_size)  return "I[" + std::to_string(reg) + "]";
  reg -= params.input_size;
  if (reg < params.nb_registers)   return "R[" + std::to_string(reg) + "]";
  reg -= params.nb_registers;
  if (reg < params.output_size) return "O[" + std::to_string(reg) + "]";
  reg -= params.output_size;
  return "M[" + std::to_string(reg) + "]";
}

void agent::print_instruction(std::ostream& os, instruction ins) const
{
#define BINARY(OP) \
  [](std::ostream& os, const agent& a, int reg_dst, int reg1, int reg2, int, int, int) \
  { \
        os \
        << a.reg_to_string(reg_dst) \
        << " <- " \
        << a.reg_to_string(reg1) \
        << OP \
        << a.reg_to_string(reg2); \
  }
#define UNARY(OP) \
  [](std::ostream& os, const agent& a, int reg_dst, int reg1, int, int, int, int) \
  { \
        os \
        << a.reg_to_string(reg_dst) \
        << " <- " \
        << OP \
        << "(" \
        << a.reg_to_string(reg1) \
        << ")"; \
  }
  static const std::function<void(std::ostream& os, const agent& a,
                                  int reg_dst, int reg1, int reg2,
                                  int opcode2,
                                  int reg3, int reg4)> print[OP_SIZE] =
    {
      [](std::ostream& os, const agent&,
         int, int, int, int, int, int)
      {
        os << "nop";
      },
      BINARY(" + "),
      BINARY(" - "),
      BINARY(" max "),
      BINARY(" min "),
      UNARY("-"),
      BINARY(" * "),
      BINARY(" / "),
      UNARY("log"),
      UNARY("exp"),
      UNARY("cos"),
      UNARY("sin"),
      UNARY("popcount"),
      [](std::ostream& os, const agent& a,
         int reg_dst, int reg1, int reg2,
         int opcode2,
         int reg3, int reg4)
      {
        os
        << "if ("
        << a.reg_to_string(reg3)
        << " < "
        << a.reg_to_string(reg4)
        << ") ";
        print[opcode2](os, a, reg_dst, reg1, reg2, 0, 0, 0);
      },
      [](std::ostream& os, const agent& a,
         int, int reg1, int reg2, int,
         int reg3, int)
      {
      // os << "error \n";
        os
        << "if ("
        << a.reg_to_string(reg1)
        << " < "
        << a.reg_to_string(reg2)
        << ") goto pc + "
        << a.reg_to_string(reg3);
        os << " ";
      }
    };
    int opcode = ins & 0xf;
    int reg_dst = (ins >> 4) & 0x7ff;
    int reg1 = (ins >> 15) & 0x7ff;
    int reg2 = (ins >> 26) & 0x7ff;
    int opcode2 = (ins >> 37) & 0xf;
    int reg3 = (ins >> 41) & 0x7ff;
    int reg4 = (ins >> 52) & 0x7ff;
    print[opcode](os, *this, reg_dst, reg1, reg2, opcode2, reg3, reg4);
}

std::ostream& operator<<(std::ostream& os, const agent& a)
{
  for (int i = 0; i < a.params.program_size; i++)
    {
    //printf("%d \n",i);
      a.print_instruction(os, a.program[i]);
      os << std::endl;
    }
  return os;
}

namespace genetic_operators
{
  /**
  * @brief Generate a random instruction
  * 
  * @param[in] a Agent in which the instruction will be inserted
  * @return The generated instruction
  */
  static agent::instruction random_instruction(const agent& a)
  {
    std::uniform_int_distribution<int> ropcode(0, agent::OP_SIZE - 1);
    std::uniform_int_distribution<int> ropcode2(1, agent::OP_IF - 1);
    std::uniform_int_distribution<int> rdestination(a.params.input_size,
                                                    a.params.input_size +
                                                    a.params.nb_registers +
                                                    a.params.output_size +
                                                    a.params.nb_long_term_registers - 1);
    std::uniform_int_distribution<int> rany(0,
                                            a.params.input_size +
                                            a.params.nb_registers +
                                            a.params.output_size +
                                            a.params.nb_long_term_registers - 1);
    // auto destination_register = [&]()
    //   {
    //     std::uniform_real_distribution<float> r(0, 1);
    //     if (r(Utils::Random::get_Rng()) < 0.15) return a.params.input_size + block;
    //     return rdestination(Utils::Random::get_Rng());
    //   };

    auto random_register = [&]()
    {
      // std::uniform_real_distribution<float> r(0, 1);
      // if (r(Utils::Random::get_Rng()) < 0.2)
      //   return destination_register();
      return rany(Utils::Random::get_Rng());
    };
    agent::instruction opcode  = ropcode(Utils::Random::get_Rng());
    agent::instruction reg_dst = rdestination(Utils::Random::get_Rng());
    agent::instruction reg1    = random_register();
    agent::instruction reg2    = random_register();
    agent::instruction opcode2 = ropcode2(Utils::Random::get_Rng());
    agent::instruction reg3    = random_register();
    agent::instruction reg4    = random_register();
    agent::instruction res     = opcode | reg_dst << 4 | reg1 << 15 | reg2 << 26
      | opcode2 << 37 | reg3 << 41 | reg4 << 52;
    return res;
  }

  void initialize(agent& a)
  {
    for (int i = 0; i < a.params.program_size * a.params.nb_parts; ++i)
      {
        a.program[i] = random_instruction(a);
      }
  }

  void mutate(agent& a)
  {
      // std::uniform_real_distribution<float> r(0, 1);
      // for (int i = 0; i < a.params.program_size; ++i)
      //   {
      //     if (r(Utils::Random::get_Rng()) < 0.05)
      //       a.program[i] = random_instruction(a);
      //   }
    std::uniform_real_distribution<float> r(0, 1);
    std::uniform_int_distribution<int> p(0, a.params.nb_parts - 1);
    int part = p(Utils::Random::get_Rng()) * a.params.program_size;
    for (int i = 0; i < a.params.program_size; ++i)
      {
        if (r(Utils::Random::get_Rng()) < 0.05)
          a.program[part + i] = random_instruction(a);
      }
  }

  void crossover(agent& a1, agent& a2)
  {
    // std::uniform_int_distribution<int> index(0, a1.params.program_size - 1);
    // int index1 = index(Utils::Random::get_Rng());
    // int index2 = index(Utils::Random::get_Rng());
    // if (index1 > index2) std::swap(index1, index2);
    // for (int i = index1; i <= index2; ++i)
    //   {
    //     std::swap(a1.program[i], a2.program[i]);
    //   }
    std::uniform_int_distribution<int> index(0, a1.params.nb_parts);
    int index1 = index(Utils::Random::get_Rng()) * a1.params.program_size;
    int index2 = index(Utils::Random::get_Rng()) * a1.params.program_size;
    if (index1 > index2) std::swap(index1, index2);
    for (int i = index1; i < index2; ++i)
      {
        std::swap(a1.program[i], a2.program[i]);
      }
  }
}
