#pragma once
#include <vector>
#include <boost/serialization/vector.hpp>

/**
* @file agent.hpp
* @brief Header of agent.cpp
*
*/

/**
* @struct agent
* @brief Represents an agent, i.e. an AI.
*
*/
struct agent{
  /**
  * @struct parameters
  * @brief Represents the caracteristics of an agent.
  *
  */
  struct parameters{
    int input_size; /**< Number of pixel composing the screen*/
    int output_size; /**< Number of possible output on the controller*/
    int nb_registers; /**< Number of register composing the register file "temporary"*/
    int nb_long_term_registers; /**< Number of register composing the register file "permanent"*/
    int program_size; /**< Total number of instruction composing a part of the program*/
    int nb_parts; /**< Number of part composing a program*/
    int time_by_part; /**< Number of time that a part is ran*/

    /**
    * @fn parameters()
    * @brief Standard constructor
    *
    */
    parameters() = default;

    /**
    * @fn parameters(int input_size, int output_size, int nb_registers, int nb_long_term_registers, int program_size, int nb_parts, int time_by_part)
    * @brief Constructor entirely initializing parameters.
    *
    * @param input_size Number of pixel composing the screen
    * @param output_size Number of possible output on the controller
    * @param nb_registers Number of register composing the register file "temporary"
    * @param nb_long_term_registers Number of register composing the register file "permanent"
    * @param program_size Total number of instruction composing a part of the program
    * @param nb_parts Number of part composing a program
    * @param time_by_part Number of time that a part is ran
    */
    parameters(int input_size, int output_size,
               int nb_registers, int nb_long_term_registers,
               int program_size, int nb_parts, int time_by_part)
      : input_size(input_size), output_size(output_size),
        nb_registers(nb_registers), nb_long_term_registers(nb_long_term_registers),
        program_size(program_size), nb_parts(nb_parts), time_by_part(time_by_part)
    {}

    template <typename Archive>
    /**
    * @fn serialize(Archive& ar, const unsigned int)
    * @brief Save/Load parameters with archive 
    *
    * @param[in,out] ar Archived parameters
    * @remark The second parameter is unused, as we don't ensure retrocompatibility.
    * @warning Don't use it. This function is automatically called when (un)serializing with boost.
    */
    void serialize(Archive& ar, const unsigned int)
    {
      ar & input_size
         & output_size
         & nb_registers
         & nb_long_term_registers
         & program_size
         & nb_parts
         & time_by_part;
    }
  };

  /**
  * @enum opcode
  * @brief Enumeration of all the operation available for an agent
  *
  */
  enum opcode
  {
    OP_NOP,
    OP_ADD,
    OP_MINUS,
    OP_MAX,
    OP_MIN,
    OP_UMINUS,
    OP_MUL,
    OP_DIV,
    OP_LOG,
    OP_EXP,
    OP_COS,
    OP_SIN,
    OP_POPCOUNT,
    OP_IF,
    OP_IF_GOTO,
    OP_SIZE
  };
  using instruction = uint64_t;
  using reg = double;
  parameters params; /**< Caracteristics of this agent*/
  std::vector<instruction> program; /**< List of instructions representing the program*/
  std::vector<reg> registers; /**< Register files of this agent*/
  int nb_steps = 0; /**< Total number of execution of the current part*/
  int current_part = 0; /**< Number corresponding to the currently executed part*/

  /**
  * @fn agent
  * @brief Standard constructor (Empty agent)
  *
  */
  agent() = default;

  /**
  * @fn agent(const parameters& params)
  * @brief Construct an agent with parameters.
  * @details Initializes params, but also the program size and the register files size. 
  *
  * @param params Parameters of this agent
  */
  agent(const parameters& params);

  /**
  * @fn get_action(const std::vector<unsigned char>& screen) 
  * @brief Execute the program on a particular frame and choose an action
  *
  * @param screen Register file corresponding to the reduced screen
  * @return The index corresponding to the highest value in register file "outputs"
  */
  int get_action(const std::vector<unsigned char>& screen);
  
  template <typename Archive>
   /**
    * @fn serialize(Archive& ar, const unsigned int)
    * @brief Save/Load agent with archive 
    *
    * @param[in,out] ar Archived agent
    * @remark The second parameter is unused, as we don't ensure retrocompatibility.
    * @warning Don't use it. This function is automatically called when (un)serializing with boost.
    */
  void serialize(Archive& ar, const unsigned int)
  {
    ar & params
       & program
       & registers;
  }

  /**
  * @fn reset_registers()
  * @brief Fills the registers "temporary" and "outputs" files with 0
  *
  */
  void reset_registers();

  /**
  * @fn reset()
  * @brief Fills all the registers files with 0 and set nb_steps and current_part to 0
  *
  */
  void reset();
  
  /**
  * @fn execute_instruction
  * @brief Perform the operation corresponding to this instruction
  *
  * @param screen The register file corresponding to the reduced screen
  * @param pc The program counter, corresponds to an instruction
  */
  inline void execute_instruction(const std::vector<unsigned char>& screen, int& pc);

  /**
  * @fn begin_outputs
  * @brief Access the beginning of the register file "outputs"
  *
  * @return An iterator pointing to the beginning of the register file "outputs"
  */
  inline auto begin_outputs();
  
  /**
  * @fn end_outputs
  * @brief Access the end of the register file "outputs"
  *
  * @return An iterator pointing to the end of the register file "outputs"
  */
  inline auto end_outputs();

  /**
  * @fn reg_to_string
  * @brief Represent a register as a string
  *
  * @param reg The index of a register
  * @return A string representing the corresponding register among all the register files
  */
  std::string reg_to_string(int reg) const;

  /**
  * @fn print_instruction
  * @brief Display an instruction as a String through a outputStream
  *
  * @param os The output stream used to display
  * @param ins The instruction to be displayed
  */
  void print_instruction(std::ostream&, instruction ins) const;
};

/**
* @fn operator<<(std::ostream&, const agent&)
* @brief Operator << overloaded so that an agent's program can be displayed by using this operator directly on the agent
*
* @return Display the agent's program
*/
std::ostream& operator<<(std::ostream&, const agent&);

/**
* @namespace genetic_operators
* @brief Handle genetic operations performed on agents
*
*/
namespace genetic_operators
{
  /**
  * @fn initialize(agent&)
  * @brief Initialize an agent with random instructions
  *
  * @param[in,out] a The agent to be initialized
  */
  void initialize(agent&);
  
  /**
  * @fn mutate(agent&)
  * @brief Mutate an agent, i.e. modify some instructions
  *
  * @param[in,out] a The agent to be affected by mutation
  */
  void mutate(agent&);
  
  /**
  * @fn crossover(agent&, agent&)
  * @brief Swap instructions between two agents
  *
  * @param[in,out] a1 The first agent involved in crossover
  * @param[in,out] a2 The second agent involved in crossover
  */
  void crossover(agent&, agent&);
}
