#pragma once

#include <boost/asio.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/archive/text_oarchive.hpp>
#include <functional>
#include <memory>
#include <iomanip>
#include <string>
#include <sstream>
#include <vector>

/**
* @file connection.hpp
* @brief Handle networking
*
*/

/**
* @class connection
* @brief Handle networking
*
*/

class connection{
  boost::asio::ip::tcp::socket socket; /**< The socket used to transfer informations*/
  constexpr static int HEADER_LENGTH = 8; /**< Length of the outputStream's header*/

  /**
  * @fn create_outbound_header(int header)
  * @brief Transform an int into his hexadecimal value. May return an error if the inner outputStream is not properly initialized.
  *
  * @param header Int value to be transformed
  * @return The hexadecimal value as a string
  */
  std::string create_outbound_header(int header){
    std::ostringstream header_stream;
    header_stream << std::setw(HEADER_LENGTH)
                  << std::hex
                  << header;
    if (!header_stream || header_stream.str().size() != HEADER_LENGTH)
      throw std::runtime_error("bad header");
    return header_stream.str();
  }
  
  public:
  /**
  * @fn connection(boost::asio::ip::tcp::socket socket)
  * @brief Standard constructor. Initialize the connection by moving a socket.
  *
  * @param socket The associated socket
  */
  connection(boost::asio::ip::tcp::socket socket)
    : socket(std::move(socket))
  {
  }

  /**
  * @fn ~connection
  * @brief Standard destructor. Close the associated socket
  *
  */
  ~connection(){
    close();
  }

  /**
  * @fn close
  * @brief Disable then close the socket. May return an error if it fails to disable the socket.
  *
  */
  void close(){
    boost::system::error_code ec;
    socket.shutdown(boost::asio::ip::tcp::socket::shutdown_both, ec);
    socket.close();
  }

  /**
  * @fn get_socket
  * @brief Return the associated socket.
  *
  * @return The associated socket.
  */
  boost::asio::ip::tcp::socket& get_socket(){
    return socket;
  }

  template <typename T>
  /**
  * @fn async_write(const T& t, boost::asio::yield_context yield)
  * @brief Write a value through the associated socket.
  *
  * @param[in] t Value to be writen
  * @param yield Context object, represents the currently executing coroutine.
  */
  void async_write(const T& t, boost::asio::yield_context yield){
    std::ostringstream archive_stream;
    boost::archive::text_oarchive archive(archive_stream);
    archive << t;
    std::string outbound_data   = archive_stream.str();
    std::string outbound_header = create_outbound_header(outbound_data.size());
    std::vector<boost::asio::const_buffer> buffers;
    buffers.push_back(boost::asio::buffer(outbound_header));
    buffers.push_back(boost::asio::buffer(outbound_data));
    boost::asio::async_write(socket, buffers, yield);
  }

  template <typename T>
  /**
  * @fn async_read(T& t, boost::asio::yield_context yield)
  * @brief Read a value through the associated socket. May return an error if the inner inputStream can't be read.
  *
  * @param[out] t Value to be read
  * @param yield Context object, represents the currently executing coroutine.
  */
  void async_read(T& t, boost::asio::yield_context yield){
    char inbound_header[HEADER_LENGTH];
    boost::asio::async_read(socket, boost::asio::buffer(inbound_header, HEADER_LENGTH), yield);
    std::istringstream is(std::string(inbound_header, HEADER_LENGTH));
    int size = 0;
    if (!(is >> std::hex >> size)){
        throw std::runtime_error("bad header");
    }
    std::vector<char> inbound_data(size);
    boost::asio::async_read(socket, boost::asio::buffer(inbound_data), yield);
    std::string archive_data(&inbound_data[0], inbound_data.size());
    std::istringstream archive_stream(archive_data);
    boost::archive::text_iarchive archive(archive_stream);
    archive >> t;
  }
};

using connection_ptr = std::shared_ptr<connection>;
