#pragma once

#include "message.hpp"
#include "master.hpp"
#include <boost/asio/spawn.hpp>
#include <boost/interprocess/sync/interprocess_semaphore.hpp>
#include <vector>
#include <memory>
#include <functional>
#include <atomic>
#include <thread>
#include <numeric>

/**
* @file fitness.hpp
* @brief Handle score computing and flow control
*
*/

/**
* @namespace genetic_algorithms
* @brief Handle score computing and flow control of the algorithm
*
*/
namespace genetic_algorithms
{
  template <typename Res, typename Params>
  /**
  * @class distributed_fitness
  * @brief Handle flow control of the algorithm
  *
  */
  class distributed_fitness{
    std::function<Res(const std::list<Res>&)> accumulate; /**< Function used to compute scores*/
    std::vector<Params> params; /**< List of agents*/
    std::vector<std::list<Res>> results; /**< List of results produced by the agents*/
    boost::interprocess::interprocess_semaphore ready; /**< Interprocess semaphore, prevent the algorithm to start a new batch of test from nowhere*/
    boost::interprocess::interprocess_semaphore processed; /**< Interprocess semaphore, prevent the algorithm to process the results if some agents are not tested yet*/
    bool stop = false; /**< Define if the algorithm is about to end*/
    std::atomic<bool> last; /**< Define if the algorithm is on the last iteration*/

  public:
    /**
    * @fn distributed_fitness(std::function<Res(const std::list<Res>&)> accumulate)
    * @brief Standard Constructor. Initialize the accumulate function, set last to false and set ready and processed to 0.
    *
    * @param accumulate Function taking a list of Res and returning a Res, in this case it computes a mean
    * @see genetic_algorithms::accumulate
    */
    distributed_fitness(std::function<Res(const std::list<Res>&)> accumulate);
    
    /**
    * @fn finished
    * @brief Get the value of stop, i.e. if the algorithm is finished or not. 
    *
    * @return The stop parameter
    */
    bool finished();
    
    /**
    * @fn get_parameters(boost::asio::yield_context yield)
    * @brief Decrements ready and get the parameters of the genetic algorithm.
    * 
    * @param yield Context object, represents the currently executing coroutine.
    * @remark yield parameter is unused
    * @return The list of parameters of the genetic algorithm
    */
    std::vector<Params> get_parameters(boost::asio::yield_context yield);

    /**
    * @fn set_results(std::vector<std::list<Res>> res, boost::asio::yield_context yield)
    * @brief Increments processed, set the results and stop the algorithm if last is true.
    *
    * @param res The list of results
    * @param yield Context object, represents the currently executing coroutine.
    */
    void set_results(std::vector<std::list<Res>> res,
                     boost::asio::yield_context yield);

    /**
    * @fn run
    * @brief Manage a batch test of agents
    *
    * @param i Current number of generation
    * @param nb_iterations Total number of generation wanted
    * @param[in] population Vector of agents to be tested
    * @param[out] results Vector of results produced by the agents
    */
    void run(int i, int nb_iterations,
             const std::vector<Params>& population,
             std::vector<Res>& results);
  };

  template <typename Res, typename Params>
  distributed_fitness<Res, Params>::distributed_fitness(std::function<Res(const std::list<Res>&)> accumulate)
    : accumulate(accumulate),
      ready(0), processed(0), last(false)
  {
  }

  template <typename Res, typename Params>
  bool distributed_fitness<Res, Params>::finished()
  {
    return stop;
  }

  template <typename Res, typename Params>
  std::vector<Params> distributed_fitness<Res, Params>::get_parameters(boost::asio::yield_context)
  {
    ready.wait();
    return params;
  }

  template <typename Res, typename Params>
  void distributed_fitness<Res, Params>::set_results(std::vector<std::list<Res>> res,
                                                     boost::asio::yield_context)
  {
    results = res;
    processed.post();
    if (last) stop = true;
  }

  template <typename Res, typename Params>
  void distributed_fitness<Res, Params>::run(int i, int nb_iterations,
                                             const std::vector<Params>& population,
                                             std::vector<Res>& res)
  {
    params.clear(); //Clear the vector of agents
    for (const auto& p : population) //Fill the vector of agents with the new population
      {
        params.emplace_back(p);
      }
    ready.post();
    if (i == nb_iterations) last = true;
    processed.wait();
    for (unsigned int i = 0; i < results.size(); ++i) //Put mean of each agent results list in a vector
      {
        res[i] = accumulate(results[i]);
      }
  }

  template <typename Res, typename Params>
  /**
  * @brief Creates a fitness tool and launches a master in a separate thread.
  *
  * @param port Port used to communicate over the network
  * @param accumulate Function used to compute score
  * @param nb_eval_by_parameter Number of test performed on an agent
  * @param nb_eval_by_slave Number max of test performed by a slave on an agent
  * @return The fitness tool
  */
  std::shared_ptr<distributed_fitness<Res, Params>> make_distributed_fitness(int port,
                                                                             std::function<Res(const std::list<Res>&)> accumulate,
                                                                             int nb_eval_by_parameter = 1, int nb_eval_by_slave = 1)
  {
    auto fitness = std::make_shared<distributed_fitness<Res, Params>>(accumulate); //Create an object distributed_fitness with accumulate as an attribute
    std::thread t([=]{ master<Res, Params>(port, nb_eval_by_parameter, nb_eval_by_slave).run(fitness); }); //Create a thread with a master and run it
    t.detach(); //Give independance to the thread
    return fitness;
  }

  /**
  * @namespace genetic_algorithms::accumulate
  * @brief Handle computation of score
  *
  */
  namespace accumulate
  {
    template <typename T>
    /**
    * @fn mean_score(const std::list<T>& l)
    * @brief Compute the mean of a list of values.
    *
    * @param[in] l List of values
    * @return Mean of the values in the parameter list
    */
    double mean_score(const std::list<T>& l)
    {
      return std::accumulate(l.begin(), l.end(), T{}, std::plus<T>()) / (double)l.size();
    }
  }
}
