#pragma once

#include <vector>
#include <utility>
#include <iostream>
#include <climits>
#include <algorithm>
#include <cassert>
#include <functional>
#include "Random.h"
#include <fstream>
#include <boost/archive/text_oarchive.hpp>
#include <boost/archive/text_iarchive.hpp>
#include <boost/serialization/vector.hpp>
#include "agent.hpp"
#include <ctime>
#include <iostream>
#include <string>
#include "Utils.hpp"
#include <sys/stat.h>
#include <sys/types.h>

/**
* @file ga.hpp
* @brief Header of ga.cpp
*
*/

/**
* @class ga
* @brief Implementation of the genetic algorithm
*
*/
class ga{
  public:
  /**
  * @struct parameters
  * @brief Parameters of the genetic algorithm
  *
  */
  struct parameters{
    double mutation_rate; /**< Percentage chance of mutation for an agent*/
    double crossover_rate; /**< Percentage chance of crossover between two agents*/
    int tournament_size; /**< Size of pools when a new population is created*/
    int population_size; /**< Size of the agent population*/
    int nb_iterations; /**< Number of times that the population is trained*/
  };
  private:
  const parameters params; /**< Structure containing the parameters*/
  std::vector<agent>  current_population; /**< Current population as a list of agents */
  std::vector<agent>  new_population; /**< Newly generated population as a list of agents*/
  std::vector<double> population_fitness; /**< Fitness score corresponding to the current population*/

  /**
  * @fn selection
  * @brief Select the best agent among a pool of randomly selected agent.
  * 
  * @return Index of a "good" agent, i.e. the best agent in a pool of randomly selected agent.
  */
  int selection() const;

  /**
  * @fn init_population
  * @brief Create whole new population with agents made of random instructions
  *
  */
  void init_population();
  
  /**
  * @fn fill_population
  * @brief Fill the population with the last population saved.
  * @remark If there is not enough agents to fill the new population, random agents are used.
  *
  */
  void fill_population();

  /**
  * @fn select_population
  * @brief Create a new population by selecting good agent
  * @see selection()
  */
  void select_population();

  /**
  * @fn crossover_population
  * @brief Perform crossover on an entire population in accordance with the crossover rate.
  *
  */
  void crossover_population();

  /**
  * @fn mutate_population
  * @brief Perform mutation on an entire population in accordance with the mutation rate.
  *
  */
  void mutate_population();

  public:
    /**
    * @fn ga(const parameters& params, const agent::parameters& agent_params)
    * @brief Standard constructor. Initializes parameters of the genetic algorithm. Also initializes vectors corresponding to the populations with the agents parameters.
    *
    * @param params Parameters of the genetic algorithm
    * @param agent_params Parameters of the agents
    */
    ga(const parameters& params, const agent::parameters& agent_params);
    
    template <typename Fitness>
    /**
    * @fn run
    * @brief Executes the genetic algorithm
    * @details First of all, the algorithm initializes the population with random agents or the last population saved.\n
    * Then, the agents are tested and their fitness calculated. If one of them is better than the current best agent, the best agent is saved.\n
    * Finally, a new population is generated by selecting best agents among pools of current agents. They are also crossed and mutated. This population is saved.\n
    * The algorithm loops on these newly generated populations until all iterations are done.\n
    * It also saves information that will be useful to produce stats as .txt file.  
    *
    * @param fresh If true, initializes the population with random agents. If false, initializes the populations with the last population saved.
    * @param fitness Tool used to processed results
    */
    void run(bool fresh, Fitness fitness);
};

template <typename Fitness>
void ga::run(bool fresh, Fitness fitness){
  std::time_t t = std::time(0); 
  std::tm* now = std::localtime(&t);
  const char* rom = ALE.romSettings->rom(); //Nom de la rom

  char aux[100];
  char best[100];
  char fichier_stat[100];
  char fichier_instr[100];
  sprintf(best,"agents/%s_%d_%d",rom,(now->tm_mon +1),(now->tm_mday));
  mkdir(best,0777);
  sprintf(fichier_stat,"%s/stat_agent_%s_%d_%d.txt",best,rom,(now->tm_mon +1),(now->tm_mday));
  sprintf(fichier_instr,"%s/instr_agent_%s_%d_%d.txt",best,rom,(now->tm_mon +1),(now->tm_mday));
  sprintf(aux,"/best_agent_%s_%d_%d.txt",rom,(now->tm_mon +1),(now->tm_mday));
  strcat(best,aux);
  printf("%s\n",best);

  std::ofstream snake; //ofstream du fichier de stat (Liste des meilleurs scores)
  std::ofstream bond; //ofstream du programme des best agent

  //Intro du fichier de stat
  snake.open(fichier_stat);
  snake << "Rom : "
      << rom
      << std::endl
      << "Mutation : "
      << params.mutation_rate
      << ", Croisement : "
      << params.crossover_rate
      << ", Taille du pool : "
      << params.tournament_size
      << ", Population : "
      << params.population_size
      << std::endl;

  if(fresh){
    init_population();
  }else{
    fill_population();
  }
  double best_value = 0;
  agent best_agent;
  for (int i = 1; i <= params.nb_iterations; ++i){
    fitness->run(i, params.nb_iterations, current_population, population_fitness);
    auto it = std::max_element(population_fitness.begin(), population_fitness.end());

    snake << *it
        << ";"
        << i
        <<std::endl;

    if (*it > best_value){
      std::cout << "ga iter: "
                << i
                << " fitness: "
                << *it
                << std::endl;
      best_value = *it;
      int best_index = std::distance(population_fitness.begin(), it);
      best_agent = current_population[best_index];
      std::ofstream ofs(best);
      boost::archive::text_oarchive oa(ofs);
      oa << best_agent;
      ofs.close();

      bond.open(fichier_instr);
      bond << best_agent;
      bond.close();
    }
    select_population();
    crossover_population();
    mutate_population();
    std::swap(current_population, new_population);
    current_population[0] = best_agent;
    std::sort(population_fitness.begin(), population_fitness.end(), std::greater<double>());
    // for (int i = 0; i < params.population_size / 10; ++i)
    //   {
    //     std::cout << population_fitness[i] << " | ";
    //   }
    // std::cout << std::endl;
    }
  snake.close();

  std::ofstream ofpop("best_pop.txt");
  boost::archive::text_oarchive opop(ofpop);
  opop << current_population;
  ofpop.close();

  auto sp = current_population.size();
  std::cout << sp << " agents archived successfully" << std::endl;
}
