#pragma once
#include <boost/asio/io_service.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/spawn.hpp>
#include <boost/chrono.hpp>
#include <iostream>
#include <memory>
#include <list>
#include <queue>
#include <vector>
#include "message.hpp"
#include "connection.hpp"

/**
* @file master.hpp
* @brief Handle the master part
*
*/

template <typename Res, typename Params>
/**
* @class master
* @brief Master part of the algorithm, handle server-side connection and redistribute evenly the tests to the slaves.
*
*/
class master
{
  int port;/**< Port used to communicate over the network*/
  int nb_eval_by_parameter;/**< Number of test performed on an agent*/
  int nb_eval_by_slave;/**< Number max of test performed by a slave on an agent*/

  boost::asio::io_service io_service; /**< Abstract networking boost interface*/
  std::list<connection_ptr> slaves; /**< List of slaves owned by this master*/
  std::queue<message::request<Params>> work_queue; /**< List of agents to be tested*/
  std::list<message::result<Res>> intermediate_results; /**< List of results produced by a batch of tested agents*/

  /**
  * @struct session
  * @brief Coroutine managing ONE slave 
  *
  */
  struct session : public std::enable_shared_from_this<session>
  {
    /**
    * @fn session::run
    * @brief Automatically launches test on agents when master is ready.
    * @details The coroutine waits for the master to fill the work queue.\n
    * Then, an agent is extracted and sent to the slave.\n
    * When the slave responds, the result is put in the list of results and the coroutine loops.\n
    * If an error occured during the agent testing, the agent is replace in the work queue.
    *
    * @param master The master corresponding to this slave
    * @param connection The slave used to performed tests
    */
    void run(master<Res, Params>& master, connection_ptr connection);

    /**
    * @fn session::close
    * @brief Remove a slave from the list of available slaves and close his connection
    *
    * @param[in] master The master corresponding to this slave
    * @param connection The slave to be removed 
    */
    void close(master<Res, Params>& master, connection_ptr connection);
  };

  /**
  * @fn init_work_queue
  * @brief Fills the work queue with the list of agent in accordance the evaluations parameters.
  * @remark This method will duplicate agents if needed (If we want to test an agent on several machines for instance).
  *
  * @param parameters_to_evaluate List of agent to be redistributed
  * @return The number of tests needed to evaluate the entire list.
  */
  unsigned int init_work_queue(const std::vector<Params>& parameters_to_evaluate);
public:
  /**
  * @fn master
  * @brief Standard constructor
  * @details Initializes port to 54321 and evaluations parameters to 1.
  *
  */
  master(int port = 54321, int nb_eval_by_parameter = 1, int nb_eval_by_slave = 1);
  
  template <typename Client>

  /**
  * @fn run
  * @brief Start a coroutine handling incoming slave connection and a coroutine handling the tests.
  *
  * @param client Tool used to process results (Fitness)
  */
  void run(Client client);
};

template <typename Res, typename Params>
void master<Res, Params>::session::run(master<Res, Params>& master, connection_ptr connection)
{
  using namespace boost;
  auto self(this->shared_from_this());

  //Coroutine de run
  asio::spawn(master.io_service,
              [self, connection, &master](asio::yield_context yield)
              {
                message::request<Params> request;
                try
                  {
                    asio::deadline_timer timer(master.io_service);
                    system::error_code ec;
                    std::string remote_address = connection->get_socket().remote_endpoint().address().to_string();
                    while (connection->get_socket().is_open()) //As long as the slave is online
                      {
                        while (master.work_queue.empty()) //Wait if we have no AI to test
                          {
                            timer.expires_from_now(posix_time::seconds(1));
                            timer.async_wait(yield[ec]);
                          }
                        request = master.work_queue.front(); //Unpack AI
                        master.work_queue.pop();
                        asio::deadline_timer timer_slave(master.io_service, posix_time::seconds(1800)); 
                        //Start ultimatum for 1800s, 
                        timer_slave.async_wait([self, connection, request, remote_address, &master](const system::error_code& ec)
                                               {
                                                 if (ec != boost::asio::error::operation_aborted) //Issue with slave, test crashed
                                                   {
                                                     master.work_queue.emplace(request);
                                                     std::cout << "abort slave from "
                                                               << remote_address
                                                               << std::endl;
                                                     self->close(master, connection);
                                                   }
                                               });
                        do //Force slave to test the AI -> If impossible the timer will make it crash anyway 
                          {
                            connection->async_write(request, yield[ec]);
                            if (ec)
                              {
                                timer.expires_from_now(posix_time::seconds(2));
                                timer.async_wait(yield[ec]);
                              }
                          } while (ec);
                        message::result<Res> res;
                        connection->async_read(res, yield[ec]); //Get the result
                        timer_slave.cancel(); //Wait for slave to finish testing
                        if (ec) //Issue with result -> re-add AI to the testList
                          {
                            std::cout << "master read rejected" << std::endl;
                            master.work_queue.emplace(request);
                            continue;
                          }
                        // std::cout << "received results from: "
                        //           << remote_address
                        //           << std::endl;
                        master.intermediate_results.emplace_back(res);
                        // if (master.intermediate_results.size() % 50 == 0)
                        //   std::cout << master.intermediate_results.size() << std::endl;
                      }
                  }
                catch (const std::exception& e)
                  {
                    master.work_queue.emplace(request);
                    std::cout << e.what() << std::endl;
                    self->close(master, connection);
                  }
              });
}

template <typename Res, typename Params>
void master<Res, Params>::session::close(master<Res, Params>& master, connection_ptr connection)
{
  master.slaves.remove(connection);
  connection->close();
}

template <typename Res, typename Params>
unsigned int master<Res, Params>::init_work_queue(const std::vector<Params>& parameters_to_evaluate)
{
  unsigned int nb_jobs = 0;
  int id = 0;
  uint64_t seed = 1;
  for (const auto& p : parameters_to_evaluate) //For each AI in params
    {
      message::request<Params> r;
      int n = std::min(nb_eval_by_parameter, nb_eval_by_slave);
      int rest = nb_eval_by_parameter;
      for (int i = 0; i < nb_eval_by_parameter; i += n, rest -= n)
        {
          r.id = id;
          r.nb_eval = n;
          r.seed = seed++;
          r.params = p;
          work_queue.emplace(r);
          ++nb_jobs;
        }
      if (rest != 0)
        {
          r.id = id;
          r.nb_eval = rest;
          r.seed = seed++;
          r.params = p;
          work_queue.emplace(r);
          ++nb_jobs;
        }
      ++id;
      seed = 1;
    }
  return nb_jobs;
}

template <typename Res, typename Params>
master<Res, Params>::master(int port, int nb_eval_by_parameter, int nb_eval_by_slave)
  : port(port), nb_eval_by_parameter(nb_eval_by_parameter), nb_eval_by_slave(nb_eval_by_slave)
{
}

template <typename Res, typename Params>
template <typename Client>
void master<Res, Params>::run(Client client)
{
  using namespace boost;
  //Coroutine de connexion
  asio::spawn(io_service,
              [&](asio::yield_context yield)
              {
                asio::ip::tcp::acceptor acceptor(io_service, asio::ip::tcp::endpoint(asio::ip::tcp::v4(), port));
                while (true)
                  {
                    system::error_code ec;
                    asio::ip::tcp::socket socket(io_service);
                    acceptor.async_accept(socket, yield[ec]);//Try to accept a connexion
                    if (!ec) //If succesfull
                      {
                        slaves.emplace_back(new connection(std::move(socket)));//Add the connection to the list of slaves
                        auto session_ptr = std::make_shared<session>();
                        std::cout << "new connection from: "
                                  << slaves.back()->get_socket().remote_endpoint().address().to_string()
                                  << std::endl;
                        session_ptr->run(*this, slaves.back());//Call run on the new slave with us as master
                      }
                  }
              });

  //Coroutine de job
  asio::spawn(io_service,
              [&](asio::yield_context yield)
              {
                system::error_code ec;
                asio::deadline_timer timer(io_service);
                while (!client->finished())//We still to iterate
                  {
                    unsigned int nb_params, nb_jobs;
                    const auto params = client->get_parameters(yield); //Get AI list
                    nb_params = params.size();
                    nb_jobs = init_work_queue(params);
                    while (intermediate_results.size() < nb_jobs)
                      {
                        timer.expires_from_now(posix_time::seconds(2));
                        timer.async_wait(yield[ec]);
                      }
                    std::cout << "master "
                              << intermediate_results.size()
                              << " jobs done"
                              << std::endl;
                    std::vector<std::list<Res>> res(nb_params);
                    for (auto& r : intermediate_results)
                      {
                        res[r.id].splice(res[r.id].end(), r.results);
                      }
                    client->set_results(std::move(res), yield);
                    intermediate_results.clear();
                  }
                io_service.stop();
              });
  io_service.run();
}
