#pragma once
#include <string>
#include <memory>
#include <functional>
#include "message.hpp"
#include <boost/asio/io_service.hpp>
#include <boost/asio/ip/tcp.hpp>
#include <boost/asio/spawn.hpp>
#include "connection.hpp"
#include <chrono>
#include <unistd.h>

/**
* @file slave.hpp
* @brief Handle the slave part
*
*/

template <typename Res, typename Params>
/**
* @class slave
* @brief Slave part of the algorithm, perform test on agents
*
*/
class slave
{
  std::string master_name; /**< Name of the machine that owned this slave*/
  std::string port; /**< Port used to communicate over the network*/
  using worker = std::function<message::result<Res>(const message::request<Params>&)>;
  worker producer; /**< Method used to test the agents*/
  int duration; /**< Slave max life-time in minutes*/

public:
  /**
  * @fn slave
  * @brief Standard constructor
  *
  * @param master_name Name of the machine that owned this slave
  * @param port Port used to communicate over the network
  * @param w Method used to test the agents
  * @param duration Slave max life-time, 1,000,000 minutes by default
  */
  slave(const std::string& master_name, int port, worker w, int duration = 1e6)
    : master_name(master_name), port(std::to_string(port)), producer(w), duration(duration)
  {
  }

  /**
  * @fn run
  * @brief In a coroutine, initialize a connection with the master throught a socket and perform as many test as requested.
  * @details First of all, a socket is created and attempt a connection to the master.\n
  * Then, when the slave receives a request, the given agent is tested and the result is sent back to the master.\n
  * The slave continues to work until the socket is disconnected or the slave duration is exceeded.
  *
  */
  void run();

  /**
  * @fn do_work
  * @brief Test an agent
  *
  * @param request Agent to be tested
  * @return Result (score) obtained by the agent
  */
  message::result<Res> do_work(const message::request<Params>&);
};

template <typename Res, typename Params>
void slave<Res, Params>::run()
{
  using namespace boost;
  asio::io_service io_service;
  asio::spawn(io_service,
              [&](asio::yield_context yield)
              {
                try
                  {
                    auto start = std::chrono::system_clock::now();
                    char machine[512];
                    gethostname(machine, 512);
                    machine[511] = 0;
                    std::time_t start_time = std::chrono::system_clock::to_time_t(start);
                    std::cout << "slave starts work on machine "
                              << machine
                              << " at time "
                              << std::ctime(&start_time)
                              << std::endl;
                    asio::ip::tcp::resolver resolver(io_service);
                    asio::ip::tcp::resolver::query query(master_name, port);
                    asio::ip::tcp::resolver::iterator endpoint_iterator = resolver.resolve(query);
                    asio::ip::tcp::socket socket(io_service);
                    asio::async_connect(socket, endpoint_iterator, yield);
                    connection connection(std::move(socket));
                    while (connection.get_socket().is_open())
                      {
                        system::error_code ec;
                        message::request<Params> request;
                        connection.async_read(request, yield[ec]);
                        if (ec)
                          {
                            std::cout << "slave can't read on machine "
                                      << machine
                                      << std::endl;
                            break;
                          }
                        message::result<Res> res = do_work(request);
                        connection.async_write(res, yield[ec]);
                        if (ec)
                          {
                            std::cout << "slave can't write on machine "
                                      << machine
                                      << std::endl;
                            break;
                          }
                        auto end = std::chrono::system_clock::now();
                        int elapsed_minutes = std::chrono::duration<double>(end - start).count() / 60.0;
                        if (elapsed_minutes >= duration)
                          {
                            std::time_t end_time = std::chrono::system_clock::to_time_t(end);
                            std::cout << "slave has finished on machine "
                                      << machine
                                      << " at time "
                                      << std::ctime(&end_time)
                                      << std::endl;
                            break;
                          }
                      }
                  }
                catch (std::exception& e)
                  {
                    std::cout << e.what() << std::endl;
                  }
              });
  io_service.run();
}

template <typename Res, typename Params>
message::result<Res> slave<Res, Params>::do_work(const message::request<Params>& request)
{
  message::result<Res> res = producer(request);
  res.id = request.id;
  return res;
}
